import cProfile
import io
import os.path
import pstats
import sys
import traceback
from math import radians
from typing import Dict, List

import bpy
from mathutils import Euler, Matrix, Quaternion, Vector

import animation as anim
import logger
import objects_custom_props as custom_props
import pyedm
from armature_export import export_armature, build_bone_id
from block_builder import BlockEnum
from edm_exception import EdmException
from enums import LampTypeEnum, NodeGroupTypeEnum, ObjectTypeEnum, EDMPropsSpecialTypeStr
from light_export import make_point_light, make_spot_light
from logger import LogCtx, log
from material_cache import MaterialCache
from materials import MATERIALS, Mat
from math_tools import ROOT_TRANSFORM_MATRIX, get_aa_bb, IDENTITY_MATRIX
from mesh_builder import buld_mesh, has_armature
from object_node import (DummyNode, LodLeaf, LodRoot, ObjectNodeCustomType, SceneRootNode)
from object_node_tree import ObjectNodeTree
from visibility_animation import extract_visibility_animation
import utils


def is_aa_bb(object: bpy.types.Object) -> bool:
    edm_props: custom_props.EDMPropsGroup = object.EDMProps
    if object.type == ObjectTypeEnum.EMPTY and edm_props.SPECIAL_TYPE in ('USER_BOX', 'BOUNDING_BOX', 'LIGHT_BOX'):
        return True
    
    return False

def is_connector(object: bpy.types.Object) -> bool:
    edm_props: custom_props.EDMPropsGroup = object.EDMProps
    if object.type == ObjectTypeEnum.EMPTY and edm_props.SPECIAL_TYPE in ('CONNECTOR'):
        return True
    
    return False

def is_mesh(object: bpy.types.Object) -> bool:
    edm_props: custom_props.EDMPropsGroup = object.EDMProps
    if object.type == ObjectTypeEnum.MESH and edm_props.SPECIAL_TYPE in ('UNKNOWN_TYPE'):
        return True
    
    return False

def is_fake_light(object: bpy.types.Object) -> bool:
    edm_props: custom_props.EDMPropsGroup = object.EDMProps
    if object.type == ObjectTypeEnum.MESH and edm_props.SPECIAL_TYPE in ('FAKE_LIGHT'):
        return True
    
    return False

def is_light(object: bpy.types.Object) -> bool:
    if object.type == ObjectTypeEnum.LIGHT or object.type == ObjectTypeEnum.LAMP:
        return True
    
    return False

def is_shell(object: bpy.types.Object) -> bool:
    edm_props: custom_props.EDMPropsGroup = object.EDMProps
    if object.type in (ObjectTypeEnum.MESH, ObjectTypeEnum.SURFACE) and edm_props.SPECIAL_TYPE in ('COLLISION_SHELL'):
        return True
    
    return False

def is_segment(object: bpy.types.Object) -> bool:
    edm_props: custom_props.EDMPropsGroup = object.EDMProps
    if object.type in (ObjectTypeEnum.MESH, ObjectTypeEnum.CURVE, ObjectTypeEnum.CURVES) and edm_props.SPECIAL_TYPE in ('COLLISION_LINE'):
        return True
    
    return False

class CollectionWalker:
    def __init__(self, context: bpy.types.Context, model: pyedm.Model) -> None:
        self.profile = cProfile.Profile()
        self.context: bpy.types.Context = context
        self.model: pyedm.Model = model
        self.material_cache = MaterialCache()
        self.obj_tree = ObjectNodeTree(context)
        self.obj_tree.build()
        self.bones = {}
        self.skins = []

    def enum_children(self, full_name, obj, node, current_armature):
        for o in obj.children:
            self.enum_object(full_name, o, node, current_armature)

    def export_aabb(self, object: bpy.types.Object):
        mat = ROOT_TRANSFORM_MATRIX @ object.matrix_world
        vec1, vec2 = get_aa_bb(mat)

        edm_props: custom_props.EDMPropsGroup = object.EDMProps

        if edm_props.SPECIAL_TYPE == 'USER_BOX':
            self.model.setUserBox((vec1.x, vec1.y, vec1.z, vec2.x, vec2.y, vec2.z))
        elif edm_props.SPECIAL_TYPE == 'BOUNDING_BOX':
            self.model.setBBox((vec1.x, vec1.y, vec1.z, vec2.x, vec2.y, vec2.z))
        elif edm_props.SPECIAL_TYPE == 'LIGHT_BOX':
            self.model.setLightBox((vec1.x, vec1.y, vec1.z, vec2.x, vec2.y, vec2.z))

    def export_light(self, object: bpy.types.Object, transform_node: pyedm.Node):
        blender_lamp: bpy.types.Light = object.data
        if blender_lamp.type == LampTypeEnum.SPOT:
            edm_light = make_spot_light(object, transform_node)
            self.model.addSpotLight(edm_light)
        elif blender_lamp.type == LampTypeEnum.POINT:
            edm_light = make_point_light(object, transform_node)
            self.model.addOmniLight(edm_light)

    def export_connector(self, object: bpy.types.Object, transform_node: pyedm.Node):
        edm_connector = pyedm.Connector(object.name)

        connector_transform = pyedm.Transform('Fake Connector Transform', Matrix.Rotation(radians(90.0), 4, 'X'))
        transform_node.addChild(connector_transform)

        edm_connector.setControlNode(connector_transform)

        edm_props: custom_props.EDMPropsGroup = object.EDMProps
        sp_scope = {}
        connector_ext: str = edm_props.CONNECTOR_EXT
        try:
            exec(connector_ext, {}, sp_scope)
        except Exception as e:
            log.error(f"Connector has {object.name} invalid format: {edm_props}. {e}")
            return
        
        for ext_name, ext_value in sp_scope.items():
            if type(ext_value) == float or type(ext_value) == int:
                edm_connector.setPropertyFloat(ext_name, float(ext_value))
            if type(ext_value) == str:
                edm_connector.setPropertyString(ext_name, ext_value)

        self.model.addConnector(edm_connector)

    def export_mesh(self, obj: bpy.types.Object, control_node: pyedm.Node, armature: bpy.types.Armature):
        mesh_storages = buld_mesh(obj, armature)
        nTriangles = 0

        edm_props = obj.EDMProps

        for mesh_storage in mesh_storages:
            material_wrap = self.material_cache.get(obj.material_slots[mesh_storage.material_index].material) if obj.material_slots else None
            if not material_wrap or not material_wrap.is_valid():
                log.warning(f"{obj.name} has no material.")
                continue

            nTriangles += mesh_storage.nTriangles
            
            mat_fx: Mat = MATERIALS.get(material_wrap.node_group_type)
            if mat_fx:
                if material_wrap.node_group_type == NodeGroupTypeEnum.DEFAULT:
                    utils.print_parents(obj)
                    edm_render_node = mat_fx.build_blocks(obj, material_wrap, mesh_storage)
                    if not edm_render_node.hasBlock(BlockEnum.BT_Bone):
                        edm_render_node.setControlNode(control_node)
                        err = self.model.addRenderNode(edm_render_node)
                        if err:
                            log.error(err)
                    else:
                        edm_bone = pyedm.Bone('Fake Control Bone ' + control_node.getName(), IDENTITY_MATRIX)
                        edm_bone.setInvertedBaseBoneMatrix(obj.matrix_local.inverted())
                        control_node = control_node.addChild(edm_bone)
                        edm_render_node.setControlNode(control_node)
                        self.skins.append(edm_render_node)
                else:
                    edm_render_node = mat_fx.build_blocks(obj, material_wrap, mesh_storage, edm_props)
                    #edm_render_node.setControlNode(fake_control_node)
                    edm_render_node.setControlNode(control_node)
                    err = self.model.addRenderNode(edm_render_node)
                    if err:
                        log.error(err)


        return (nTriangles, control_node)
    
    def export_fake_light(self, obj: bpy.types.Object, control_node: pyedm.Node, armature: bpy.types.Armature):
        bpy_mesh: bpy.types.Mesh = obj.data
        nLights = len(bpy_mesh.vertices)
        material_wrap = self.material_cache.get(obj.material_slots[0].material) if obj.material_slots else None
        if not material_wrap or not material_wrap.is_valid():
             log.fatal(f"{obj.name} has no material.")
        
        mat_fx: Mat = MATERIALS.get(material_wrap.node_group_type)
        if mat_fx:
            edm_render_node = mat_fx.build_blocks(obj, material_wrap, None)
            edm_render_node.setControlNode(control_node)
            err = self.model.addRenderNode(edm_render_node)
            if err:
                log.error(err)

        return (nLights, control_node)
    
    def export_shell(self, obj: bpy.types.Object, control_node: pyedm.Node):
        mesh_storages = buld_mesh(obj, None)
        nTriangles = 0
        for mesh_storage in mesh_storages:
            nTriangles += mesh_storage.nTriangles
            
            edm_shell_node = pyedm.ShellNode(obj.name)
            edm_shell_node.setIndices(mesh_storage.indices)
            edm_shell_node.setPositions(mesh_storage.positions)
            edm_shell_node.setControlNode(control_node)
            self.model.addShellNode(edm_shell_node)

        return (nTriangles, control_node)

    def enum_object(self, parent_object_path: str, obj: ObjectNodeCustomType, edm_parent_node: pyedm.Node, current_armature: bpy.types.Armature):
        logger.LOG_CTX.obj = obj
        try:
            full_name = os.path.join(parent_object_path, obj.name)

            if type(obj) is DummyNode:
                edm_node = edm_parent_node.addChild(pyedm.Node(obj.name))
                self.enum_children(full_name, obj, edm_node, current_armature)
                return
            
            if type(obj) is SceneRootNode:
                edm_node = edm_parent_node.addChild(pyedm.Node(obj.name))
                self.enum_children(full_name, obj, edm_node, current_armature)
                return

            if type(obj) is LodRoot:
                levels = [x.dist for x in obj.children]
                edm_node = edm_parent_node.addChild(pyedm.Lod(obj.name, levels))
                self.enum_children(full_name, obj, edm_node, current_armature)
                return
            
            if type(obj) is LodLeaf:
                edm_node = edm_parent_node.addChild(pyedm.Node(obj.name))
                self.enum_children(full_name, obj, edm_node, current_armature)
                return

            if not obj.visible:
                self.enum_children(full_name, obj, edm_parent_node, current_armature)
                return

            vis_node: pyedm.VisibilityNode = extract_visibility_animation(obj.obj)
            if vis_node:
                edm_parent_node = edm_parent_node.addChild(vis_node)

            edm_node = anim.extract_transform_animation(edm_parent_node, obj.obj)
            if obj.obj.parent_bone:
                pbone = obj.obj.parent_bone
                bn = self.bones[build_bone_id(current_armature.name, pbone)]
                tn = pyedm.Transform(f'End Of {bn.name}', Matrix.LocRotScale((0, bn.bone.length ,0), None, None))
                ebn = bn.edm_bone.addChild(tn)
                edm_node = ebn.addChild(edm_node)
                log.debug(f'{obj.obj.parent_bone} --> {obj.obj.name}')
            
            edm_props: custom_props.EDMPropsGroup = obj.obj.EDMProps
            if is_light(obj.obj):
                self.export_light(obj.obj, edm_node)
                log.info(f"{full_name} as {obj.obj.type} {edm_props.SPECIAL_TYPE}")
            elif is_aa_bb(obj.obj):
                self.export_aabb(obj.obj)
                log.info(f"{full_name} as {obj.obj.type} {edm_props.SPECIAL_TYPE}")
            elif is_connector(obj.obj):
                self.export_connector(obj.obj, edm_node)
                log.info(f"{full_name} as {obj.obj.type} {edm_props.SPECIAL_TYPE}")
            elif is_fake_light(obj.obj):
                nLights, edm_node = self.export_fake_light(obj.obj, edm_node, current_armature)
                log.info(f"{full_name} as {obj.obj.type}. Control node: {edm_node.getName()}. N triangles: {nLights}.")
            elif is_mesh(obj.obj):
                nTriangles, edm_node = self.export_mesh(obj.obj, edm_node, current_armature)
                log.info(f"{full_name} as {obj.obj.type}. Control node: {edm_node.getName()}. N triangles: {nTriangles}.")
            elif is_shell(obj.obj):
                if has_armature(obj.obj.modifiers):
                    log.fatal(f"Shell {obj.name} has bones.")
                nTriangles, edm_node = self.export_shell(obj.obj, edm_node)
                log.info(f"{full_name} as {obj.obj.type} {nTriangles}")
            elif obj.obj.type == ObjectTypeEnum.ARMATURE:
                current_armature = obj.obj
                edm_node = export_armature(obj.obj, edm_node, self.bones)
                log.info(f"{full_name} as {obj.obj.type}")
            else:
                log.info(f"{full_name} as {obj.obj.type}")
                    
            self.enum_children(full_name, obj, edm_node, current_armature)

        except EdmException as e:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            res = ''.join(traceback.format_tb(exc_traceback, limit=1))
            res += ''.join(traceback.format_exception(exc_type, exc_value, exc_traceback, limit=2))
            log.error(str(e), res)
            
    def build_skin(self):
        for skin in self.skins:
            bb = skin.getBlock(BlockEnum.BT_Bone)
            bnames = bb.getBoneNames()
            bone_objs = []

            for b in bnames:
                # As name of bone in skin isn't real it's name of vertex group.
                if b in self.bones:
                    bone_objs.append(self.bones[b].edm_bone)
                else:
                    log.fatal(f"Bone '{b}' doesn't exist, please remove vertex group with the same name.")
            bb.setBones(bone_objs)

            err = self.model.addRenderNode(skin)
            if err:
                log.error(f"Can't export skin node {skin.getName()}. Reason: {err}")

    def do(self) -> None:
        self.profile.enable()
        root = pyedm.Transform('', ROOT_TRANSFORM_MATRIX)
        #root = pyedm.Transform('', IDENTITY_MATRIX)
        self.model.getRootTransform().addChild(root)
        self.enum_object('', self.obj_tree.obj_tree, root, None)
        self.build_skin()
        self.profile.disable()

    def log_status(self) -> None:
        ios = io.StringIO()
        ps = pstats.Stats(self.profile, stream = ios).sort_stats(pstats.SortKey.CUMULATIVE)
        ps.print_stats()
        #print(ios.getvalue())

def _write(context: bpy.types.Context, edm_file_path: str) -> bool:
    logger.LOG_CTX = LogCtx()
    model = pyedm.Model()
    try:
        walker = CollectionWalker(context, model)
        walker.do()
        walker.log_status()

        if not log.errors:
            result_str: str = model.save(edm_file_path, 10)
            if result_str:
                log.fatal(f"Can't save model {edm_file_path}. Reason: {result_str}")
    except Exception as e:
        logger.LOG_CTX = None
        raise e
    finally:
        del model
        del walker
        nAlived = pyedm.get_num_alived_objects()
        if nAlived:
            log.warning(f"{nAlived} objects are still alive.")

    
    logger.LOG_CTX = None