import numpy as np
from armature_export import build_bone_id
from logger import log

class MeshStorage:
    def __init__(self, nTriangles, uvNames, uv_active, material_index, armature) -> None:
        self.cur = 0
        self.nVerts = 0
        self.material_index = material_index
        self.armature = armature
        self.nTriangles = nTriangles
        self.positions = np.empty(nTriangles * 9, dtype=np.float32)
        self.normals = np.empty(nTriangles * 9, dtype=np.float32)
        self.indices = np.empty(nTriangles * 3, dtype=np.uint32)
        self.indices_map = {}
        self.uv_active = uv_active
        self.uv = {name: np.empty(nTriangles * 6, dtype=np.float32) for name in uvNames}

        self.bones = {}
        self.bone_indices = np.empty(nTriangles * 12, dtype=np.uint32)
        self.bone_weights = np.empty(nTriangles * 12, dtype=np.float32)
        self.cur_bone_index = 0
        self.is_skin = False

    def prepare(self, vertices, bverts, vertex_groups, orig_normals, orig_uvs, vertices_indices):
        self.vertices = vertices
        self.bverts = bverts
        self.vertex_groups = vertex_groups
        self.orig_normals = orig_normals
        self.orig_uvs = orig_uvs
        self.vertices_indices = vertices_indices
        if bverts:
            self.is_skin = True

    def set(self, loops):
        for index in loops:
            vertex_index = self.vertices_indices[index]

            if (vertex_index, index) in self.indices_map:
                i = self.indices_map[(vertex_index, index)]
                self.indices[self.cur] = i
                self.cur += 1
                continue

            if self.is_skin:
                weights_temp = [0.0, 0.0, 0.0, 0.0]
                temp_inds = [0, 0, 0, 0]

                j = 0
                for gr in self.bverts[vertex_index // 3].groups:
                    if gr.weight < 0.0001:
                        continue

                    if j > 3:
                        log.fatal("Skin vertex has more than 4 infuencing bones.")

                    grpid = gr.group

                    bone_name = build_bone_id(self.armature.name, self.vertex_groups[grpid].name)
                    if bone_name not in self.bones:
                        self.bones[bone_name] = self.cur_bone_index
                        self.cur_bone_index += 1

                    ind = self.bones[bone_name]
                    weights_temp[j] = gr.weight
                    temp_inds[j] = ind
                    j = j + 1
     
                c = self.nVerts * 4
                self.bone_indices[c : c + 4] = temp_inds
                self.bone_weights[c : c + 4] = weights_temp

            i = len(self.indices_map)
            self.indices_map[(vertex_index, index)] = i
            self.indices[self.cur] = i

            c = self.nVerts * 3
            self.positions[c : c + 3] = self.vertices[vertex_index : vertex_index + 3]
            self.normals[c : c + 3] = self.orig_normals[index * 3 : index * 3 + 3]

            c = self.nVerts * 2
            for k in self.orig_uvs.keys():
                #self.uv[k][c : c + 2] = self.orig_uvs[k][index * 2 : index * 2 + 2]
                if not len(self.orig_uvs[k]) == 0:
                    self.uv[k][c : c + 2] = self.orig_uvs[k][index * 2 : index * 2 + 2]
                else:
                    self.uv[k][c : c + 2] = 0.0
            self.cur += 1
            self.nVerts += 1

    def shrink(self):
        self.nTriangles = self.cur // 3
        self.positions = self.positions[:self.nVerts * 3]
        self.normals = self.normals[:self.nVerts * 3]
        self.indices = self.indices[:self.nTriangles * 3]
        self.bone_indices = self.bone_indices[:self.nVerts * 4]
        self.bone_weights = self.bone_weights[:self.nVerts * 4]
        for k in self.uv.keys():
            self.uv[k] = self.uv[k][:self.nVerts * 2]